/*
 * ecibsd.c
 *
 * Upload firmware and start synchronization
 *
 *
 * Changes :
 *
 * [Boris HUISGEN - bhuisgen@wanadoo.fr]
 *    - BSD port from Linux driver
 *    - fix main to start only synchonization if modem already exist. 
 *    - command line changed. You must now pass option -v for show infos
 *    - synchro timeout managed with signal SIGALRM : we restart the synchro
 *      if SYNCHRO_TIMEOUT delay exceeded (default fixed to 30 seconds)
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "pusb.h"
#include "modem.h"

#define FIRMWARE_TIMEOUT    1000
#define DSL_TIMEOUT         2000
/* synchro timeout in seconds */
#define SYNCHRO_TIMEOUT     30

int verbose = 0;
char **my_argv;
pid_t child_pid;
pusb_device_t dev;
pusb_endpoint_t ep;

struct firm_block
{
	unsigned short	addr;
	unsigned short	len;
	unsigned char	*content;
};

struct usb_block
{
	unsigned char   request_type;
	unsigned char   request;
	unsigned short  value;
	unsigned short  index;
	unsigned short  size;
	unsigned char	*buf;
};


void
print_char (unsigned char c)
{
    if ((c >= ' ') && (c < 0x7f))
		printf ("%c", c);
	else
		printf (".");
}

void
dump (unsigned char* buf, int len)
{
    int i, j;
  
	for (i = 0; i < len; i += 16)
	{
		for (j = i; (j < len) && (j < i + 16); j++)
			printf ("%02x ", buf[j]);
			
		for (; j < i + 16; j++)
			printf ("   ");
			
		for (j = i; j < len && j < i + 16; j++)
			print_char (buf[j]);
		
		printf("\n");
	}
}

int
firm_block_read (FILE *fp, struct firm_block *p)
{
	unsigned char b[4];
	int r;

	if ((r = fread (b, 1, sizeof (b), fp)) != sizeof (b))
	{
		printf ("firm_block_read : read %d bytes instead of %d\n", r,
		   sizeof (b));
		return (0);
	}

	p->addr = (b[0] << 8) | b[1];
	p->len  = (b[2] << 8) | b[3];

	p->content = (unsigned char *) realloc (p->content, p->len);
	if (p->content == NULL)
	{
		perror ("malloc");
		return (0);
	}

	if ((r = fread (p->content, 1, p->len, fp)) != p->len)
	{
		printf ("firm_block_read : read %d bytes instead of %d\n", r, p->len);
		return (0);
	}

	return (1);
}

int
usb_block_read(FILE *fp, struct usb_block *p)
{
	unsigned char b[8];
	int r;

	if ((r = fread (b, 1, sizeof (b), fp)) != sizeof (b))
	{
		printf ("usb_block_read: read %d bytes instead of %d\n", r, sizeof (b));
		return (0);
	}

	p->request_type = b[0];
	p->request      = b[1];
	p->value        = (b[2]<<8) | b[3];
	p->index        = (b[4]<<8) | b[5];
	p->size         = (b[6]<<8) | b[7];

	if (p->size != 0)
	{
		p->buf = (unsigned char *)realloc (p->buf, p->size);
		if (p->buf == NULL)
		{
			printf ("usb_block_read: can't allocate %d bytes\n", p->size);
			return (0);
		}

		if ((p->request_type & 0x80) == 0)
		{
			if ((r = fread (p->buf, 1, p->size, fp)) != p->size)
			{
				printf ("usb_block_read: read %d bytes instead of %d\n",
				    r, p->size);
				return (0);
			}
		}
	}

	return (1);
}

void
read_endpoint (pusb_device_t dev, int epnum)
{
	unsigned char lbuf[0x40];
	int ret;
	int pkt16_received = 0;

	fflush(stdout);
	fflush(stderr);

	child_pid = fork ();
	if (child_pid != 0)
		return ;

	ep = pusb_endpoint_open (dev, epnum, O_RDONLY);
	if (ep == NULL)
	{
		perror ("pusb_endpoint_open");
		_exit (-1);
	}

	for (;;)
	{
		ret = pusb_endpoint_read_int (ep, lbuf, sizeof (lbuf));
		if (ret < 0)
		{
			printf ("Error reading interrupts\n");
			break;
		}
		
		if (verbose)
		{
			printf ("endpoint %02x, packet len = %d\n", epnum, ret);
	       
		    dump (lbuf,ret);
		    fflush (stdout);
		    fflush (stderr);
   		}

		if ((epnum == 0x86) && (ret != 64))
			kill (getppid(), SIGUSR1);

		if (ret == 16)
			pkt16_received = 1;

		if ((pkt16_received) && (ret == 64))
			break;
	}

	pusb_endpoint_close (ep);
	
	_exit (0);
}

int
check_modem (unsigned short vid2, unsigned short pid2, int retry)
{
#define RETRY_PERIOD 10
    struct timeval tv, start;
	int i, r;

	retry++;

	gettimeofday (&start,NULL);
	sleep (1);

	for (i = 0; i < retry; i++)
	{
		fd_set rset;
		FD_ZERO(&rset);

		dev = pusb_search_open (vid2, pid2);
		if (dev != NULL)
			break;

		tv.tv_sec = RETRY_PERIOD / 1000;
		tv.tv_usec = (RETRY_PERIOD % 1000) * 1000;

		r = select (1, NULL, NULL, NULL, &tv);
		if (r < 0)
		{
		   perror ("select");
		   break;
		}  
	}

	if (dev == NULL)
		return (0);
	
	pusb_close (dev);
	
	return (1);
}

int
load_firmware (const char * file, unsigned short vid1, unsigned short pid1)
{
	FILE *fp;
	struct firm_block block;
	long size;
	int n = 0;

	fp = fopen (file, "rb");
	if (fp == NULL)
	{
		perror(file);
		return (0);
	}

	fseek (fp, 0, SEEK_END);
	size = ftell (fp);
	fseek (fp, 0, SEEK_SET);

	dev = pusb_search_open (vid1, pid1);
	if (dev == NULL)
	{
		printf ("load_firmware : can't find modem\n");
		fclose (fp);
		return (0);
	}

	/* init device */
	if (pusb_set_configuration (dev, 1) < 0)
	{
		printf ("load_firmware : can't set configuration 1\n");
		pusb_close (dev);
		fclose (fp);
		return (0);
	}

	if (pusb_set_interface (dev, 0, 1) < 0)
	{
		perror ("load_firmware : can't set interface 0 to use alt setting 1\n");
		pusb_close (dev);
		fclose (fp);
		return (0);
	}

	block.content = NULL;

	while (ftell (fp) < size)
	{
		n++;

		if (!firm_block_read (fp, &block))
		{
			printf("can't read block.\n");
			pusb_close (dev);
			fclose (fp);
			
			return (0);
		}
		
		if (verbose)
		   printf ("block %3d : Addr 0x%04x - Length %2u : ", n, block.addr,
		      block.len);
  
		if (pusb_control_msg (dev, 0x40, 0xa0, block.addr, 0, block.content,
		      block.len, FIRMWARE_TIMEOUT) < 0)
		{
		   if (verbose)
		      printf ("failed\n");
		   
		   pusb_close (dev);
		   fclose (fp);
		   
		   return (0);
		}
        if (verbose)
           printf("ok\n");
	}

	pusb_close (dev);
	fclose (fp);

	return (1);
}

int
load_synchro (const char * file, unsigned short vid2, unsigned short pid2)
{
	FILE *fp;
	struct usb_block block;
	long size ;
	int n = 0, r, i;

	fp = fopen (file, "rb");
	if (fp == NULL)
	{
		perror(file);
		return 0;
	}

	fseek (fp ,0, SEEK_END);
	size = ftell (fp);
	fseek (fp, 0, SEEK_SET);

	dev = pusb_search_open (vid2, pid2);
	if (dev == NULL)
	{
		printf ("load_synchro : can't find modem\n");
		fclose (fp);
		return (0);
	}

	/* init device */
	if (pusb_set_configuration (dev,1) < 0)
	{
		printf("load_synchro : can't set configuration 1\n");
		pusb_close (dev);
		fclose (fp);
		return (0);
	}

	sleep(1);
	
	/* warning : orginal setting is 0,4 (source : Windows driver) */
	if (pusb_set_interface (dev, 0, 4) < 0)
	{
		printf ("load_synchro : can't set interface 0 to use alt setting 4\n");
		pusb_close (dev);
		fclose (fp);
		return (0);
	}

    alarm (SYNCHRO_TIMEOUT);
	
	read_endpoint (dev, 0x86);

	memset(&block, 0, sizeof(block));

	while (ftell (fp) < size)
	{
		n++;

		if (!usb_block_read (fp, &block))
		{
			printf("can't read block\n");
			pusb_close (dev);
			fclose (fp);

			return (0);
		}
		
		if (verbose)
		   printf ("Block %3d : request_type=%02x request=%02x value=%04x index=%04x size=%04x : ",
		      n, block.request_type, block.request, block.value, block.index,
		      block.size);
		  
		if ((r = pusb_control_msg (dev, block.request_type, block.request,
		          block.value, block.index, block.buf,
		          block.size, DSL_TIMEOUT)) != block.size)
		{
		   if (verbose)
		      printf ("Failed r=%d\n", r);
             
		   pusb_close (dev);
		   fclose (fp);
		   return (0);
		}

		if ((r != block.size) && verbose)
			printf("(expected %d, got %d) ",block.size,r);

		r = block.size;
 
		if (verbose)
		{
		   printf("Ok\n");
		   
		   if ((block.request_type & 0x80))
		   {
		      for (i = 0; i < r; i++)
		         printf("%02x ", block.buf[i]);
        	  
              printf("\n");
           }
		}

		if ((block.request_type == 0x40) && (block.request == 0xdc) &&
		    (block.value == 0x00) && (block.index == 0x00))
		{
			struct timeval tv1, tv2;

			if (verbose)
			{
			   printf("waiting ... ");
			   
			   fflush(stdout);
			   fflush(stderr);
			   
			   gettimeofday(&tv1, NULL);
			}
   
			/* pause() should be unblock by our son sending*/
			pause ();
			
			if (verbose)
			{
			   gettimeofday (&tv2, NULL);
			   printf ("waited %ld ms\n", (long)
			      (((tv2.tv_sec - tv1.tv_sec) * 1000) +
			       ((tv2.tv_usec - tv1.tv_usec) / 1000) ));
            }
		}

		if (verbose)
		{
		   fflush(stdout);
		   fflush(stderr);
		}  

	}
	
	alarm (0);
	
	pusb_close(dev);

	if (ftell (fp) != size)
		printf("file size = %ld <=> bytes read = %ld\n", size, ftell (fp));

	fclose (fp);

	return (1);
}

void
sigusr1 ()
{
   if (verbose)
      printf ("get SIGUSR1\n");
}

void
sigalrm ()
{
    /* we restart yourself */
    printf ("Synchronization timeout.\n");
    
    /* we kill child process */
    kill (child_pid, SIGKILL);
    /* and we cleanup device for ugen */
    pusb_endpoint_close (ep);
    pusb_close (dev);
    
    sleep (1);
    
    if (execv (my_argv[0], my_argv) == -1)
       printf ("Failed to restart.\n");    

    _exit (0);
}

void
usage ()
{
	printf ("ecibsd\n");
	printf ("usage: ecibsd [OPTIONS] -firm <file> -synchro <file>\n");
	printf ("   -vid1 <num>       chip VID, default set to 0x547\n");
	printf ("   -pid1 <num>       chip PID, default set to 0x2131\n");
	printf ("   -vid2 <num>       modem VID, default set to 0x0915\n");
	printf ("   -pid2 <num>       modem PID, default set to 0x8000\n");
	printf ("   -v                verbose output\n");
	printf ("   -firm <file>      binary file of chip's firmware\n");
	printf ("   -synchro <adsl>   binary file of synchronization\n");
	_exit (0);
}

int
main (int argc, char *argv[])
{
    unsigned short vid1, pid1, vid2, pid2;
    const char *firm;
    const char *synchro;
    int i, r, status;
    
    /* keep arguments */
    my_argv = argv;
	
    /* set default values */
    verbose = 0;
    vid1 = strtoul ("0x0547", NULL, 0);
    pid1 = strtoul ("0x2131", NULL, 0);
    vid2 = strtoul ("0x0915", NULL, 0);
    pid2 = strtoul ("0x8000", NULL, 0);
    firm = NULL;
    synchro = NULL;
    	
    /* parsing command line */	
    for (i = 1; i < argc; i++)
    {
       if (strcmp(argv[i], "-vpi1") == 0 && i + 1 < argc)
	  vid1 = strtoul (argv[++i], NULL, 0);
       else if (strcmp(argv[i], "-vci1") == 0 && i + 1 < argc)
	  pid1 = strtoul (argv[++i], NULL, 0);
       else if (strcmp(argv[i], "-vpi2") == 0 && i + 1 < argc)
	  vid2 = strtoul (argv[++i], NULL, 0);
       else if (strcmp(argv[i], "-vci2") == 0 && i + 1 < argc)
	  pid2 = strtoul (argv[++i], NULL, 0);
       else if (strcmp(argv[i], "-v") == 0)
	  verbose = 1;
       else if (strcmp(argv[i], "-firm") == 0 && i + 1 < argc)
	  firm = argv[++i];
       else if (strcmp(argv[i], "-synchro") == 0 && i + 1 < argc)
	  synchro = argv[++i];
       else if (strcmp(argv[i], "--help") == 0)
	  usage();
       else 
	  usage();
    }

    if (!firm || !synchro)
      usage ();

    if (!check_modem (vid2, pid2, 0))
    {
       printf ("Loading firmware...\n");    
       if (!load_firmware (firm, vid1, pid1))
       {
          printf ("Failed to load firmware.\n");
          return (-1);
       }

       sleep (4);

       if (!check_modem (vid2, pid2, 10))
       {
          printf ("Modem not found.\n");
          return (-1);
       }
    }

    signal (SIGUSR1, sigusr1);
    signal (SIGALRM, sigalrm);

    printf ("Synchronizing line...\n");
    if (!load_synchro(synchro, vid2, pid2))
    {
       printf ("Failed to synchronize line.\n");
       return (-1);
    }
    
    while ((r = wait (&status)) > 0)
       ;

    if (errno != ECHILD)
       perror ("wait");

    printf ("\nModem ready.\n");

    return (0);
}

