/*
 * pppoeci.c
 *
 * Manage modem dialog between USB and PPP
 *
 *
 * Changes :
 *
 * [Boris HUISGEN - bhuisgen@wanadoo.fr]
 *    - BSD port from Linux driver
 *    - fix open flags for endpoints
 *    - don't exec init_ep_88 (EP_DATA_IN)
 *    - fork endpoints 88 and 86, and poll them for data
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/resource.h>
#ifdef USE_SYSLOG
#include <syslog.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <dirent.h>
#include <termios.h>
#include <poll.h>
#include <string.h>
#include <limits.h>
#include <stdarg.h>
#include <pthread.h>

#include "pusb.h"
#include "modem.h"
#include "crc.h"
#include "smallsem.h"

/* max size of data endpoint (Windows driver uses 448). */
#define MAX_EP_SIZE         448
/* number of simultaneous URB submitted to data endpoint (Windows drv : 20 */
#define NB_PKT_EP_DATA_IN   20
/* number of ISO packet per URB (Windows driver uses 10) */
#define PKT_NB              5
/* handle ep_int */
#define MAX_BUFFER 20

/* AAL5 */
#define PAD_SIZE            11
#define CELL_SIZE           53
#define CELL_HDR            05
#define CELL_DATA           48

/* log semaphore key*/
#define	LOG_SEMKEY          0xCAFEDECA
/* USB timeout */
#define DATA_TIMEOUT        1000
/* size of hdlc header frame */
#define HDLC_HEADER_SIZE    2
/* PPP level */
#define syncHDLC            1

/* report flags*/
   /* actions */
#define REPORT_DATE         ((unsigned int)0x00000001)
#define REPORT_PERROR       ((unsigned int)0x00000002)
#define REPORT_DUMP         ((unsigned int)0x00000004)
   /* message types */
#define REPORT_ERROR        ((unsigned int)0x00000010)
#define REPORT_INFO         ((unsigned int)0x00000020)
#define REPORT_DEBUG        ((unsigned int)0x00000040)


/* VPI & VCI settings */
static int my_vpi = -1;
static int my_vci = -1;
/* verbosity level */
static int verbose = 0;
/* log */
static int log_mutex = 0;
/* process id */
static pid_t this_process;
static pid_t parent_process;
static pid_t handler88_process;
static pid_t handler02_process;
/* I/O fd */
static int gfdout;

static pusb_device_t fdusb;
static pusb_endpoint_t ep_data_out;
static pusb_endpoint_t ep_data_in;
static pusb_endpoint_t ep_int;

/* predeclarations */
int aal5_read (unsigned char *cell_buf, size_t count, unsigned int vpi,
	       unsigned int vci, unsigned int pti, int fdout);

/*******************************************************************************
 *
 * Log functions
 *
 *******************************************************************************/

unsigned char
give_char (unsigned char c)
{
	if ((c >= ' ') && (c < 0x7f))
		return (c);
	else
		return ('.');
}

void
dump (unsigned char *buf, int len)
{
	int i, j;

#ifdef USE_SYSLOG
	unsigned char line[128];
	unsigned char *ptr;
#endif

	if ((buf == NULL) || (len == 0))
		return;

	for (i = 0; i < len; i += 16)
	{
#ifdef USE_SYSLOG
		ptr = line;
#endif
		for (j = i; (j < len) && (j < i + 16); j++)
		{
#ifndef USE_SYSLOG
			printf ("%02x ", buf[j]);
#else
			sprintf (ptr, "%02x ", buf[j]);
			ptr += 3;
#endif
		}
		
		for (; j < i + 16; j++)
		{
#ifndef USE_SYSLOG
			printf ("   ");
#else
			sprintf (ptr, "   ");
			ptr += 3;
#endif
		}
		
		for (j = i; j < len && j < i + 16; j++)
		{
#ifndef USE_SYSLOG
			printf ("%c", give_char(buf[j]));
#else
			sprintf (ptr, "%c", give_char(buf[j]));
			ptr++;
#endif
		}
#ifndef USE_SYSLOG
		printf ("\n");
#else
		*ptr = '\0';
		syslog (LOG_DEBUG, line);
#endif
	}
	
#ifndef USE_SYSLOG
	printf("\n");
#endif
}

void
report (int minlevel, unsigned int flags, char *format, ...)
{
    va_list ap;
	int length = 0;
	char *buffer = NULL;

	if (verbose >= minlevel)
	{
		sem_P (log_mutex);
		
		va_start (ap, format);
#ifndef USE_SYSLOG
		if (flags & REPORT_DATE)
		{
			time_t tps;
			time (&tps);
			printf ("[%.24s] ", ctime (&tps));
		}
		
		if (flags & REPORT_ERROR)
			printf ("Error ");
		else if (flags & REPORT_INFO)
			printf ("Info  ");
		else if (flags & REPORT_DEBUG)
		    printf ("Debug ");
		
		if (flags)
			printf (">%d< ", (int)this_process); 
#endif
		if (!(flags & REPORT_DUMP))
		{
#ifndef USE_SYSLOG
			vprintf(format, ap);
#else
			char msg[256];
			
			vsnprintf (msg, 256, format, ap);
			syslog (LOG_INFO, msg);
#endif
		}
		else
		{
#ifdef USE_SYSLOG
			unsigned char msg[256];
#endif
			buffer = va_arg (ap, char*);
			length = va_arg (ap, int);

#ifndef USE_SYSLOG
			vprintf(format, ap);
#else
			vsnprintf (msg, 256, format, ap);
			if (flags & REPORT_ERROR)
				syslog (LOG_ERR, msg);
			else if (flags & REPORT_INFO)
				syslog (LOG_INFO, msg);
			else if (flags & REPORT_DEBUG)
				syslog (LOG_DEBUG, msg);
#endif
		}

#ifndef USE_SYSLOG
		if (flags & REPORT_PERROR)
			printf ("Reason : %s\n", strerror (errno));
#endif
		if (flags & REPORT_DUMP)
			dump (buffer, length);

		va_end (ap);
		sem_V (log_mutex);
	}

	return;
}


/*******************************************************************************
 *
 * PPP I/O functions
 *
 *******************************************************************************/

int
ppp_read (int fd, unsigned char **buf, int *n)
{
	static unsigned char ppp_buf[64 * 1024];
 	int r;

	for (;;)
	{
		do
		{
			r = read (fd, ppp_buf, sizeof (ppp_buf));
		}
		while ((r < 0) && (errno == EINTR));
		
		if (r < 0)
		{
			report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
				"Error reading from pppd\n");
			return (-1);
		}
		
		report (2, REPORT_DEBUG | REPORT_DATE, " Read from ppp, len = %d\n", r);
		if (verbose > 1)
			dump (ppp_buf, r);
		
		if (r <= 0)
			return (0);
		
		if (syncHDLC)
		{
			if (r < 3)
			{
				report (0, REPORT_ERROR | REPORT_DATE,
					"Read from ppp short on data, r=%d => ignored\n", r);
				continue;
			}
			
			if (r == sizeof (ppp_buf))
			{
				report (0, REPORT_ERROR | REPORT_DATE,
					"Read from ppp too much data, r=%d => ignored\n", r);
				continue;
			}
			
			*buf = ppp_buf + HDLC_HEADER_SIZE;
			*n = r - HDLC_HEADER_SIZE;
			
			return (r - HDLC_HEADER_SIZE);
		}
		else
			report (0, REPORT_ERROR | REPORT_DATE, "async HDLC not handled\n");
		
		break;
	}
}

#define NOBUF_RETRIES 5
int
ppp_write (int fd, unsigned char *buf, int n)
{
	unsigned char ppp_buf[64 * 1024];
	int r;
	static int errs = 0;
	int retries = 0;

	if (syncHDLC)
	{
		ppp_buf[0] = 0xff;	//FRAME_ADDR;
		ppp_buf[1] = 0x03;	//FRAME_CTRL;
		
		memcpy (ppp_buf + HDLC_HEADER_SIZE, buf, n);
		
		if (verbose > 1)
		{
			report (2, REPORT_DEBUG | REPORT_DATE,
				" writing to ppp, len = %d\n", n + HDLC_HEADER_SIZE);
			dump (ppp_buf, n + HDLC_HEADER_SIZE);
		}
		
		retry:
		
		if ((r = write (fd, ppp_buf, n + HDLC_HEADER_SIZE)) < 0)
		{
			if (errno == ENOBUFS)
			{
				errs++;
				
				if ((errs < 10) || (errs % 100 == 0))
					report (0, REPORT_ERROR | REPORT_DATE,
						"ppp_write: %d ENOBUFS errors\n", errs);
				
				if (retries++ < NOBUF_RETRIES)
				{
					sleep (1);
					goto retry;
				}
				else
					return (0);
			}
			else
				report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
					"Error writing to ppp(d)\n");
		}
		
		return (r);
	}
	else
		report (0, REPORT_ERROR | REPORT_DATE, "async HDLC not handled\n");
	
	return (-1);
}

/*******************************************************************************
 *
 * Cell & AAL5 functions
 *
 *******************************************************************************/

int
cell_read (unsigned char *cellbuf, size_t count, int fdout)
{
	unsigned int vpi, vci, pti;

	vpi = ( cellbuf[0]         <<  4) | (cellbuf[1] >> 4);
	vci = ((cellbuf[1] & 0x0f) << 12) | (cellbuf[2] << 4) | (cellbuf[3] >> 4);
	pti = ( cellbuf[3] & 0x0f);

	return (aal5_read (cellbuf + CELL_HDR, CELL_DATA, vpi, vci, pti, fdout));
}

int
cell_write (unsigned char *cellbuf, unsigned char *buf, int n,
    unsigned int vpi, unsigned int vci, unsigned int pti)
{
	unsigned char gfc = 0;
	unsigned char hec = 0x00;

	if (n != CELL_DATA)
		return(-1);

	vpi &= 0xff;
	vci &= 0xffff;
	pti &= 0x0f;

	cellbuf[0] = (gfc << 4) | (vpi >> 4);
	cellbuf[1] = (vpi << 4) | (vci >> 12);
	cellbuf[2] = (vci >> 4);
	cellbuf[3] = ((vci & 0xf) << 4) | pti;
	cellbuf[4] = hec;

	memcpy (cellbuf + CELL_HDR,  buf, CELL_DATA);
	memset (cellbuf + CELL_SIZE,0xff, PAD_SIZE);

	return (CELL_SIZE + PAD_SIZE);
}

int
aal5_read (unsigned char *cell_buf, size_t count, unsigned int vpi,
    unsigned int vci, unsigned int pti, int fdout)
{
	static unsigned char buf[64 * 1024];
	static int len = 0;

	int ret;
	unsigned int real_len;
	unsigned int crc1, crc2;

	if ((vpi != my_vpi) || (vci != my_vci))
	   return (0);

	if (len + CELL_DATA < sizeof (buf))
	{
		memcpy (buf + len, cell_buf, CELL_DATA);
		len += CELL_DATA;
	}
	else
	{
		report (0, REPORT_ERROR | REPORT_DATE,
		   "aal5_read: buffer overflow (len=%d) => reseting buffer\n", len);
		len = 0;
		return (0);
	}

	if ((pti & 2) == 0)
		return (0);

	real_len = (buf[len - 6] << 8) | buf[len - 5];

	report (1, REPORT_DEBUG | REPORT_DATE,
	   "< pid=%d > aal5_read: len=%d => real_len=%d\n", this_process, len,
	   real_len);

	crc1 = ~aal5_calc_crc (buf, len-4, -1);
	crc2 = (buf[len-4] << 24) | (buf[len-3] << 16) |
             (buf[len-2]<<8) | buf[len-1];

	if ((crc1 != crc2) && !(verbose > 1 ))
	   report (0, REPORT_ERROR | REPORT_DATE,
	      "aal5_read: len=%d => real_len=%d\n", len, real_len);

		
	if (crc1 != crc2)
	{
		report (0, REPORT_ERROR | REPORT_DATE,
			"bad crc %08x (instead of %08x) => dropping AAL5 frame\n", crc2,
		    crc1);
		
		len = 0;
		
		return (0);
	}

	ret = ppp_write (fdout, buf, real_len);
	len = 0;
	if (ret < 0)
	{
		report (0, REPORT_ERROR | REPORT_DATE,
			"Error writing to pppd : %d => exit\n", ret);
		_exit (0);
	}

	return (ret);
}

int
aal5_write (pusb_endpoint_t ep_data_out, unsigned char *buf, int n)
{
	/* add the trailer */
	int total_len = 48 * ((n + 8 + CELL_DATA - 1) / 48);
	unsigned int crc, pti = 0;

	/*
	* this buffer will contain all cells for an AAL5 frame,
	* so AAL5 frame is max 64K + 48 that's 1367 cells at max
	*/

	unsigned char bigbuf[1367 * 53];
	int ptr;
	int ret;

	memset(buf + n, 0, total_len - n);

	buf[total_len - 6] = n >> 8;
	buf[total_len - 5] = n & 0xff;

	crc = ~aal5_calc_crc(buf, total_len - 4, -1);

	buf[total_len - 4] = crc >> 24;
	buf[total_len - 3] = crc >> 16;
	buf[total_len - 2] = crc >> 8;
	buf[total_len - 1] = crc;

	ptr = 0;
	while (total_len > 0)
	{
		if (total_len <= CELL_DATA)
			/* this is the last cell, so mark it */
		    pti = 2;
		
		ptr += cell_write (bigbuf + ptr, buf, CELL_DATA, my_vpi, my_vci, pti);
		
		/* send the same cell over another vpi/vci */
		buf += CELL_DATA;
		total_len -= CELL_DATA;
	}
	
	if (verbose > 0)
	{
		report (1, REPORT_DEBUG | REPORT_DATE, " writing to usb, len = %d\n",
			ptr);
		dump (bigbuf, ptr);
    }

	ret = pusb_endpoint_write_bulk (ep_data_out, bigbuf, ptr, DATA_TIMEOUT);
	if (ret < 0)
		report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
			"Error writing to usb, see the reason\n");
	
	return (ret);
}

void
replace_b1_b2 (unsigned char *b1, unsigned char *b2)
{
	/* we should return :
	   0x11 0x11 0x13 0x13 0x13 0x13 0x11 0x11 0x01 0x01 0x01 0x01
	   0x11 0x11 0x13 0x13 0x13 0x13 0x11 0x11 0x53 0x53 0x53 0x53 
	   The last line is repeated infinitely
	*/	
	static int count = 0;
	static unsigned char replace_b1[] = {
		0x73, 0x73, 0x63, 0x63, 0x63, 0x63, 0x73, 0x73, 0x63, 0x63, 0x63, 0x63
	};
	static unsigned char replace_b2[] = {
		0x11, 0x11, 0x13, 0x13, 0x13, 0x13, 0x11, 0x11, 0x01, 0x01, 0x01, 0x01
	};

	*b1 = replace_b1[count];
	*b2 = replace_b2[count];

	count = (count + 1) % 12;

	if (count == 0)
		replace_b2[8] = replace_b2[9] = replace_b2[10] = replace_b2[11] = 0x53;
}

void
decode_endpoint_86 (unsigned char * buf, int size, pusb_device_t fdusb)
{
	unsigned char outbuf[40] = {
		0xff, 0xff, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c,  0x0c, 
		0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c,  0x0c, 
		0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c,  0x0c, 
		0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c, 0x0c,  0x0c, 
		0x0c, 0x0c
	};

	int i, outi = 0;
	static int lost_synchro = 0;

	if (verbose > 1)
	{
		report (1, REPORT_DEBUG | REPORT_DATE, " read from ep int, len = %d\n",
		   size);
		dump (buf, size);
	}

	if (!lost_synchro)
	{
		for (i = 3; i < 18; i++)
		{
			unsigned short w = (buf[2*i+0] << 8) | buf[2*i+1];
			
			if (w!=0x0c0c &&
				w!=0x734d && w!=0x7311 && w!=0xf301 && w!=0xf34f)
			{
				report (0, REPORT_ERROR | REPORT_DATE,
				   "lost synchronization ! w=%04x\n",w);
				lost_synchro = 1;
			}
		}
	}
	
	for (i = 0; i < 15; i++)
	{
		unsigned char b1, b2;

		b1 = buf[6+2*i+0];
		b2 = buf[6+2*i+1];

		/* each word != 0x0c 0x0c is copied in the output buffer */
		if (b1 != 0x0c || b2 != 0x0c)
		{
			/* however 0x7311 need to be replace by
			   0x6313, 0x6301, 0x6313, 0x6353 ...  */
			if (b1 == 0x73 && b2 == 0x11)
				replace_b1_b2 (&b1,&b2);


			/* we check that we are not writing outside our buffer.
			   From what is found in our log, this never happen.
			*/
			if (10 + outi >= sizeof (outbuf))
			{
				report (0, REPORT_ERROR | REPORT_DATE,
				   "Warning: outbuf is too small\n");
				dump (buf,size);				
				break;
			}

			outbuf[10 + outi++] = b1;
			outbuf[10 + outi++] = b2;
		}
	}

	if (outi != 0)
	{
		if (pusb_control_msg (fdusb, 0x40, 0xdd, 0xc02, 0x580, outbuf,
		        sizeof (outbuf), DATA_TIMEOUT) != sizeof (outbuf))
		{
			report (1, REPORT_ERROR | REPORT_DATE,
			   "Error on sending vendor device URB\n");
		}
		else
		{
			if (verbose > 1)
			{
				report (1, REPORT_ERROR | REPORT_DATE,
				   " ctrl msg sent, len = %d\n", sizeof(outbuf));
				dump (outbuf, sizeof(outbuf));
			}
		}
	}
}

int
decode_endpoint_88 (unsigned char *buf, int len)
{
	static unsigned char rbuf[64*1024];
	static int rlen = 0;
	int r, pos;

	if (verbose > 1)
	{
		report (1, REPORT_DEBUG | REPORT_DATE, " read from usb, len = %d\n",
			len);
		dump (buf, len);
	}

	/* add the received buf to our buf, if possible */
	if ((unsigned int)rlen + len < sizeof (rbuf))
	{
		memcpy (rbuf + rlen, buf, len);
		rlen += len;
	}
	else
		report (0, REPORT_ERROR | REPORT_DATE,
			"warning: no space for %d bytes\n", len);

	pos = 0;
	while ((rlen - pos) >= CELL_SIZE)
	{
		if (((r = cell_read (rbuf + pos, CELL_SIZE, gfdout)) < 0) && (r != -2))
		{
			rlen -= pos;
			if (rlen != 0)
				memcpy(rbuf, rbuf + pos, rlen);
			
			return (r);
		}
		
		if (r == -2)
			pos++;
		else
			pos += CELL_SIZE;
	}
	
	rlen -= pos;

	if (rlen != 0)
		memcpy (rbuf, rbuf + pos, rlen);

	return (0);
}

/*******************************************************************************
 *
 * Endpoints functions
 *
 *******************************************************************************/

int
init_endpoint_86 (pusb_endpoint_t ep_int)
{
#define NB_PKT_EP_INT 64
  static unsigned char buf[NB_PKT_EP_INT][0x40];
  int ret, i;
  
	for (i = 0; i < NB_PKT_EP_INT; i++)
	{
		ret = pusb_endpoint_read_int (ep_int, buf[i], sizeof (buf[i]));
		if (ret < 0)
		{
			report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
				"Error in initial submit of URB %d on EP_INT\n", i);
			return (ret);
		}
	}

	return (ret);
}

int
init_endpoint_88 (pusb_endpoint_t ep_data_in)
{    
  static unsigned char buf[NB_PKT_EP_DATA_IN][MAX_EP_SIZE*PKT_NB];
  int i, ret;

	for (i = 0; i < NB_PKT_EP_DATA_IN; i++)
	{
		ret = pusb_endpoint_read_iso (ep_data_in, buf[i],
				MAX_EP_SIZE*PKT_NB);
		if (ret < 0)
		{
			report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
				"Error on initial submit of <<URB>> %d on EP_DATA_IN\n", i);
			return (ret);
		}
	}

	return (ret);
}

void
handle_endpoint_86 (pusb_endpoint_t ep_int, pusb_device_t fdusb)
{
	struct pollfd pfd[1];
	unsigned char buf[MAX_BUFFER];
	int size, ret;
	
	memset (&pfd, 0, sizeof (struct pollfd));
	pfd[1].fd = pusb_endpoint_get_descriptor (ep_int);
	pfd[1].events = POLLIN;
	
	for (;;)
	{
		ret = poll (pfd, 1, INFTIM);	  
		if (ret == -1)
			break;
		   		
		if (!(pfd[0].revents & POLLIN))
			continue;
		
		while ((size = pusb_endpoint_read_int (ep_int, buf, sizeof (buf))) > 0)
			decode_endpoint_86 (buf, size, fdusb);
	}
	
	report (2, REPORT_DEBUG | REPORT_DATE, "handle_endpoint_86 : exiting\n");
}

void
handle_endpoint_88 (pusb_endpoint_t ep_data_in)
{
	pid_t child_pid;
	
	if ((child_pid = fork ()) == 0)
	{
		struct pollfd pfd[1];
		unsigned char buf[MAX_EP_SIZE];
		int size, ret;
		
		this_process = handler88_process = getpid();
		report(0, REPORT_INFO | REPORT_DATE,
		       "handle_endpoint_88 : child process %d\n",
		       this_process);
		
		memset (&pfd, 0, sizeof (struct pollfd));
		pfd[0].fd = pusb_endpoint_get_descriptor (ep_data_in);
		pfd[0].events = POLLIN;
		
		for (;;)
		{
			ret = poll (pfd, 1, INFTIM);
			if (ret == -1)
				break;
			
			if (!(pfd[0].revents & POLLIN))
				continue;
			
			while ((size = pusb_endpoint_read_int (ep_data_in, buf,
   						sizeof (buf))) > 0)
			  decode_endpoint_88 (buf, size);
		}
		
		report (2, REPORT_ERROR | REPORT_DATE,
			"handle_endpoint_88 : exit process\n");
		_exit (0);
	}
	else
		handler88_process = child_pid;
}

void
handle_endpoint_02 (pusb_endpoint_t ep_data_out, int fdin)
{
	pid_t child_pid;
	
	if ((child_pid = fork ()) == 0)
	{
		int r, n;
		unsigned char *buf;
		
		this_process = handler02_process = getpid();
		report(0, REPORT_INFO | REPORT_DATE,
		       "handle_endpoint_02 : child process %d\n",
		       this_process);
		
		for (;;)
		{
			n = ppp_read (fdin, &buf, &n);
			if (n <= 0)
			{
				report (2, REPORT_ERROR | REPORT_DATE, "ppp_read=%d\n", n);
				break;
			}
			
			r = aal5_write (ep_data_out, buf, n);
			if (r < 0)
			{
				report (2, REPORT_ERROR | REPORT_DATE, "aal5_write=%d\n",r);
				break;
			}
		}
		
		report (2, REPORT_ERROR | REPORT_DATE,
			"handle_endpoint_02 : exit process\n");
		_exit (0);
	}
	else
		handler02_process = child_pid;
}

/*******************************************************************************
*
* Main function
*
*******************************************************************************/

void
sighandler (int signal)
{
	report (0, REPORT_INFO | REPORT_DATE, "Killed by a signal %d\n", signal);

	// cleanup
	if (kill (handler88_process, SIGKILL) < 0)
		report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
					"Error killing handler88 child process %d\n",
					handler88_process);
					
	if (kill (handler02_process, SIGKILL) < 0)
		report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
					"Error killing handler02 child process %d\n",
					handler02_process);
					
	pusb_endpoint_close (ep_int);
	pusb_endpoint_close (ep_data_in);
	pusb_endpoint_close (ep_data_out);
	pusb_close (fdusb);
   	
	sem_destroy (log_mutex);

	_exit (1);
}

void
usage ()
{
	fprintf (stderr, "pppoeci\n");
#ifndef USE_SYSLOG
	fprintf (stderr, "usage: pppoeci [-v level] [-f logfile] -vpi val -vci val\n");
#else
	fprintf (stderr, "usage: pppoeci [-v level] -vpi val -vci val\n");
#endif
	fprintf (stderr, "  -v       : defines the verbosity level [0-2]\n");
#ifndef USE_SYSLOG
	fprintf (stderr, "  -f       : defines the log filename to use (default /var/log/pppoeci.log)\n");
#endif
	fprintf (stderr, "  -vpi.vci : defined the vpi.vci that your provider is\n");
	fprintf (stderr, "             using, 8.35 for France, 0.38 for UK\n");
	fprintf (stderr, "  -d device: defines the device to use\n");
	fprintf (stderr, "  --help   : this message\n");
	exit (-1);
}

int main(int argc, char *argv[])
{
 	int fdin, fdout;
	int i;
#ifndef USE_SYSLOG
	time_t ourtime;
	char *logfile;
	int log;
#endif
	char *devicename = NULL;

#ifndef USE_SYSLOG
	logfile = "/var/log/pppoeci.log";
	log = 0;
#endif

    /*
     * parsing command line
     */
	for (i = 1; i < argc; i++)
	{
		if (strcmp (argv[i], "-v") == 0 && i + 1 < argc)
			verbose = atoi(argv[++i]);
		else if (strcmp (argv[i], "-vpi") == 0 && i + 1 < argc)
			my_vpi = atoi(argv[++i]);
		else if (strcmp (argv[i], "-vci") == 0 && i + 1 < argc)
			my_vci = atoi(argv[++i]);
		else if (strcmp (argv[i], "-d") == 0 && i + 1 < argc)
			devicename = argv[++i];
		else if (strcmp (argv[i], "--help") == 0)
			usage();
#ifndef USE_SYSLOG
		else if (strcmp(argv[i], "-f") == 0 && i + 1 < argc)
			logfile = argv[++i];
#endif
		else 
			usage();
	}

	if ((my_vpi == -1) || (my_vci == -1))
		usage();

	fdin = dup (STDIN_FILENO);
	fdout = dup (STDOUT_FILENO);

	parent_process = this_process = getpid();
 	
	/*
	 * syslog support
	 */
#ifndef USE_SYSLOG
	log = open (logfile, O_CREAT | O_RDWR | O_APPEND, 0600);
	if (log < 0)
	{
		perror (logfile);
		return (-1);
	}
#else
	openlog ("pppoeci", LOG_PID, LOG_DAEMON);
#endif

    /*
     * initialize semaphore
     */
	if ((log_mutex = sem_get ((key_t)LOG_SEMKEY)) != -1)
	{
		if (sem_destroy (log_mutex) == -1)
		{
#ifndef USE_SYSLOG
			perror ("sem_destroy");
#else
			syslog (LOG_ERR, "sem_destroy");
#endif
			return (-1);
		}
	}
	if ((log_mutex = sem_create ((key_t)LOG_SEMKEY)) == -1)
	{
#ifndef USE_SYSLOG
		perror ("sem_create");
#else
		syslog (LOG_ERR, "sem_create");
#endif
		return (-1);
	}
	if (sem_init (log_mutex, 1) == -1)
	{
#ifndef USE_SYSLOG
		perror ("sem_init");
#else
		syslog (LOG_ERR, "sem_init");
#endif
		return (-1);
	}
	
	/*
	 * redirect all streams
	 */
#ifndef USE_SYSLOG
	if (dup2 (log, STDIN_FILENO) == -1)
	{
		write(log,"Error duplicating STDIN_FILENO\n", 31);
		return (-1);
	}
	if (dup2 (log, STDOUT_FILENO) == -1)
	{
		write(log,"Error duplicating STDOUT_FILENO\n", 32);
		return (-1);
	}
	if (dup2 (log, STDERR_FILENO) == -1)
	{
		write(log,"Error duplicating STDERR_FILENO\n", 32);
		return (-1);
	}

	if (close (log) == -1)
	{
		fprintf (stderr, "Error closing log fd\n");
		return (-1);
	}
	/* no buffering, direct flush */
	setbuf (stdout, NULL);
	setbuf (stderr, NULL);
#endif

    /*
     * begin report session
     */
	report (0, REPORT_INFO, "Starting pppoeci2\n");
#ifndef USE_SYSLOG
	time (&ourtime);
	report (0, REPORT_INFO, "Log started on %s\n", ctime(&ourtime));
#endif
	report (2, REPORT_INFO, "Using vpi=%d, vci=%d\n", my_vpi, my_vci);
	report (1, REPORT_DEBUG, "pty descriptors : fdin=%d, fdout=%d\n",
	   fdin, fdout);

	/*
	 * Increase priority of the process
	 */
	if (setpriority (PRIO_PROCESS, this_process, -20) < 0)
		report (1, REPORT_INFO | REPORT_DATE | REPORT_PERROR,
			"setpriority failed\n");
	      
    /*
     * We change the pppd send & receive buffer for FreeBSD
     */
#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
#define SOCKBUF (64*1024)
	{
		int sbuf, ss = sizeof (sbuf);

		if (getsockopt (fdin, SOL_SOCKET, SO_SNDBUF, &sbuf, &ss) == 0)
		{
			report (1, REPORT_INFO | REPORT_DATE,
			   "Increasing SNDBUF from %d to %d\n", sbuf, SOCKBUF);

			sbuf = SOCKBUF;

			if (setsockopt (fdin, SOL_SOCKET, SO_SNDBUF, &sbuf, ss) < 0)
				report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
				   "setsockopt failed\n");
       	}

		if (getsockopt (fdin, SOL_SOCKET, SO_RCVBUF, &sbuf, &ss) == 0)
		{
			report (1, REPORT_INFO | REPORT_DATE,
			   "Increasing RCVBUF from %d to %d\n", sbuf, SOCKBUF);

			sbuf = SOCKBUF;

			if (setsockopt (fdin, SOL_SOCKET, SO_RCVBUF, &sbuf, ss) < 0)
				report (0, REPORT_ERROR|REPORT_DATE|REPORT_PERROR,
				   "setsockopt failed\n");
		}
	}
#endif

	/*
	 * Install HDLC line discipline on fdin if it is a tty
	 */
	if (isatty (fdin))
	{
#ifdef N_HDLC
		int disc = N_HDLC;
		
		if (ioctl (fdin, TIOCSETD, &disc) < 0)
		{
			report (0, REPORT_ERROR | REPORT_DATE, "Error loading N_HDLC\n");
			return (-1);
		}
		
		report (2, REPORT_INFO | REPORT_DATE, "N_HDLC line set up\n");
#elif defined TTYDISC
		int disc = TTYDISC;
		
		if (ioctl (fdin, TIOCSETD, &disc) < 0)
		{
			report (0, REPORT_ERROR | REPORT_DATE,
			   "Error setting termios tty line discipline\n");
			return (-1);
		}
		
		report (2, REPORT_INFO | REPORT_DATE, "TTYDISC line set up\n");
#endif
	}
	
	/*
     * search and open device
     */
    if (devicename == NULL)
    {
		fdusb = pusb_search_open (GS_VENDOR, GS_PRODUCT);
		if (fdusb == NULL)
		{
			report (0, REPORT_ERROR | REPORT_DATE, "Modem not found\n");
			return (-1);
		}
	}
	else
	{
		fdusb = pusb_open (devicename);
		if (fdusb == NULL)
		{
			report (0, REPORT_ERROR | REPORT_DATE, "The modem is not at %s\n",
			   devicename);
			return (-1);
		}
	}

	report (2, REPORT_DEBUG | REPORT_DATE, "Modem found\n");
	gfdout = fdout;
	
	/*
	 * claim interface 1
	 */
	if (pusb_claim_interface (fdusb, 0) < 0)
	{
		report (0, REPORT_ERROR | REPORT_DATE,
		   "pusb_claim_interface 1 failed\n");
		return (-1);
	}
	
	/*
	 * open endpoints 0x86, 0x86 and 0x02
	 */
	ep_int = pusb_endpoint_open (fdusb, EP_INT, O_RDONLY);
	if (ep_int == NULL)
	{
	   report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
	      "pusb_endpoint_open failed for endpoint INT\n");
	   return (-1);	
	}
	
	ep_data_in = pusb_endpoint_open (fdusb, EP_DATA_IN, O_RDONLY);
	if (ep_data_in == NULL)
	{
	   report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
	      "pusb_endpoint_open failed for endpoint DATA_IN\n");
	   return (-1);	
	}

	ep_data_out = pusb_endpoint_open (fdusb, EP_DATA_OUT, O_WRONLY);
	if (ep_data_out == NULL)
	{
		report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
		   "pusb_endpoint_open failed for endpoint DATA_OUT\n");
		return (-1);
	}
	
    /*
     * initialize endpoints (cleanup)
     */    
	if (init_endpoint_86 (ep_int) < 0)
	  return (-1);
		
#if defined(__linux__)
	if (init_endpoint_88 (ep_data_in) < 0)
	  return (-1);
#endif
	
	/*
	 * endpoints 0x02 and 0x88 have them own thread
	 */
	handle_endpoint_02 (ep_data_out, fdin);
	handle_endpoint_88 (ep_data_in);
 
  	/*
	 * on Linux, pppd send SIGTERM for kill our process.
	 * on BSD, this is a SIGHUP 
	 */
#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
    signal (SIGHUP, sighandler);
#elif defined(__linux__)
	signal (SIGTERM, sighandler);
#endif 

	/*
	 * this process handles endpoint 0x86
	 */	
	handle_endpoint_86 (ep_int, fdusb);	
	
    /*
     * cleanup
     */
	if (kill (handler88_process, SIGKILL) < 0)
		report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
					"Error killing handler88 child process %d\n",
					handler88_process);
					
	if (kill (handler02_process, SIGKILL) < 0)
		report (0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
					"Error killing handler02 child process %d\n",
					handler02_process);

 	if (pusb_release_interface (fdusb, 0) < 0)
		report(0, REPORT_ERROR | REPORT_DATE | REPORT_PERROR,
					"pusb_release_interface failed\n");

	pusb_endpoint_close (ep_int);
	pusb_endpoint_close (ep_data_in);
	pusb_endpoint_close (ep_data_out);
	pusb_close (fdusb);
	
	sem_destroy (log_mutex);

	return (0); 
}

