/*
 * pusb.c
 *
 * USB library - BSD version
 *
 *
 * Changes :
 *
 * [Boris HUISGEN - bhuisgen@wanadoo.fr]
 *    - specific functions for bulk/iso/int endpoints
 *    - pusb_endpoint_get_descriptor()
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <poll.h>
#include <errno.h>
#include <string.h>
#include <dev/usb/usb.h>
#include <sys/ioctl.h>
#include <sys/param.h>

#include "pusb.h"

#define MAX_CONTROLLERS 10

/******************************************************************************
* Patch USB structures for the change in early 2002 
******************************************************************************/
#if ((defined (__FreeBSD__)) && (__FreeBSD_version > 500000 && __FreeBSD_version < 500031) || (__FreeBSD_version < 450001)) || (defined (__NetBSD__) && (__NetBSD_version__ < 105280000)) || (defined (__OpenBSD__))

/* struct usb_ctl_request */
#define ucr_actlen  actlen
#define ucr_flags   flags
#define ucr_data    data
#define ucr_request request
#define ucr_addr    addr

/* struct usb_alt_interface */
#define uai_alt_no          alt_no
#define uai_interface_index interface_index

/* struct usb_device_info */
#define udi_productNo productNo
#define udi_product   product
#define udi_vendorNo  vendorNo
#define udi_vendor    vendor
#define udi_devnames  devnames
#define udi_addr      addr
#endif


struct pusb_device_t {
    char *prefix;
    int fd;
    int timeout;
};

struct pusb_endpoint_t {
	int fd;
	int timeout_read;
	int timeout_write;
};

/*****************************************************************************
*	Local Prototypes
*****************************************************************************/

static pusb_device_t make_device (int fd, const char *prefix);
static pusb_endpoint_t make_endpoint (int fd);

/*****************************************************************************
*	Library functions
*****************************************************************************/

static pusb_device_t make_device (int fd, const char *prefix)
{
	pusb_device_t dev;

	if (!(dev = malloc (sizeof (*dev))))
		return (0);

	dev->fd = fd;
	dev->timeout = -1;
	if (!(dev->prefix = strdup (prefix)))
	{
		free (dev);
		return (0);
	}

	return (dev);
}

static pusb_endpoint_t make_endpoint (int fd)
{
	pusb_endpoint_t ep;

	if (!(ep = malloc (sizeof (*ep))))
	  return (0);

	ep->fd = fd;
	ep->timeout_write = -1;
	ep->timeout_read = -1;

	return (ep);
}

pusb_device_t pusb_search_open (int vendorID, int productID)
{
	int controller, device;
	int cfd, dfd;
	char buf[20], prefix[20];

	for (controller = 0; controller < MAX_CONTROLLERS; controller++)
	{
		sprintf (buf, "/dev/usb%d", controller);
		cfd = open (buf, O_RDONLY);
		if (cfd < 0)
		{
			if ((errno != ENXIO) && (errno != ENOENT))
      			fprintf(stderr, "pusb_search_open: can't open %s: %s\n",
      			         buf, strerror(errno));

			continue;
		}
		
		for (device = 1; device < USB_MAX_DEVICES; device++)
		{
			struct usb_device_info di;
			
			di.udi_addr = device;
			if (ioctl(cfd, USB_DEVICEINFO, &di) < 0)
				continue;
#ifdef debug
			{
				int j;
				fprintf (stderr, "something on controller %d device %d\n",
				    controller, device);
				fprintf (stderr, "vendor %s product %s\n", di.udi_vendor,
				    di.udi_product);
    
				for (j = 0; j < 4; j++)
				    fprintf (stderr, "name %s\n", di.devnames[j]);
			}
#endif
			if ((di.udi_vendorNo == vendorID) && (di.udi_productNo == productID))
			{
			  if (strncmp (di.udi_devnames[0], "ugen", 4) != 0)
			  {
			        /* Has a real driver, don't mess with it */
			        continue;
   			  }
   			   
   			  sprintf (prefix, "/dev/%s", di.udi_devnames[0]);

#if defined(__FreeBSD__)
			  sprintf(buf, "%s", prefix);
#else		
			  sprintf(buf, "%s.00", prefix);
#endif
              dfd = open (buf, O_RDWR);
              
              if (dfd < 0)
              {
                 fprintf (stderr, "pusb_search_open: couldn't open device %s: %s\n",
                    buf, strerror (errno));
                 close(cfd);
                 
                 return (0);
              }
              else
              {
                 close(cfd);
                 
                 return (make_device (dfd, prefix));
              }
			}
		}
	}

#ifdef debug
	fprintf (stderr, "didn't find device\n");
#endif
	close (cfd);
	
	return (0);
}

pusb_device_t pusb_open (const char *path)
{
	int dfd;
	char buf[20], prefix[20];
	
	sprintf (prefix, "%s", path);
#if defined(__FreeBSD__)
	sprintf (buf, "%s", prefix);
#else		
	sprintf (buf, "%s.00", prefix);
#endif
	dfd = open (buf, O_RDWR);
	
	if (dfd < 0)
	{
		fprintf (stderr, "pusb_search_open: couldn't open device %s: %s\n", buf,
          strerror(errno));
		  
		return (0);
	}
	else
		return (make_device (dfd, prefix));
}

int pusb_close(pusb_device_t dev)
{
	int ret;
	
	ret = close (dev->fd);
	free (dev->prefix);
	free (dev);
	
	return (ret);
}

/* Function not existing in this USB stack */
int pusb_claim_interface(pusb_device_t dev,int interface)
{
	return 0;
}

/* Function not existing in this USB stack */
int pusb_release_interface(pusb_device_t dev,int interface)
{
	return 0;
}

int pusb_control_msg (pusb_device_t dev, int request_type, int request,
    int value, int index, unsigned char *buf, int size, int timeout)
{
	struct usb_ctl_request req;
	
	req.ucr_request.bmRequestType = request_type;
	req.ucr_request.bRequest = request;
	USETW (req.ucr_request.wValue, value);
	USETW (req.ucr_request.wIndex, index);
	USETW (req.ucr_request.wLength, size);
	
	req.ucr_data = buf;
	req.ucr_flags = USBD_SHORT_XFER_OK;
	
	if (ioctl (dev->fd, USB_DO_REQUEST, &req) < 0)
		return (-1);
	else
		return (req.ucr_actlen);
}

int pusb_set_configuration (pusb_device_t dev, int config)
{
	return (ioctl (dev->fd, USB_SET_CONFIG, &config));
}

int pusb_get_configuration (pusb_device_t dev)
{
	int config;
	
	if (ioctl (dev->fd, USB_GET_CONFIG, &config) < 0)
		return (-1);
	
	return (config);
}

int pusb_set_interface (pusb_device_t dev, int interface, int alternate)
{
	struct usb_alt_interface intf;
	int config;
	
	intf.uai_interface_index = interface;
	intf.uai_alt_no = alternate;
	
	if (ioctl (dev->fd, USB_SET_ALTINTERFACE, &intf) < 0)
	{
		perror("USB_SET_ALTINTERFACE");
		return (-1);
	}
	
	/* BUG: at least in FreeBSD 4.3, USB_SET_ALTINTERFACE does not
	 * make the necessary devices for the interface.  So we call
	 * USB_SET_CONFIG to ensure they are created.  But this will
	 * fail if any endpoints (other than the control) are open.
	 * Of course, the devices might already exist, so failure
	 * to set the configuration does not mean it will not work.
	 */
	config = pusb_get_configuration (dev);
	if ((config < 0) || (pusb_set_configuration(dev, config) < 0))
		fprintf (stderr, "warning: pusb_set_interface may have failed\n");
	
	return (0);
}

int pusb_get_interface (pusb_device_t dev, int interface)
{
	struct usb_alt_interface intf;
	
	intf.uai_interface_index = interface;
	
	if (ioctl (dev->fd, USB_GET_ALTINTERFACE, &intf) < 0)
		return (-1);

	return (intf.uai_alt_no);
}

pusb_endpoint_t pusb_endpoint_open (pusb_device_t dev, int epnum, int flags)
{
	char buf[20];
	int fd;

#if defined(__FreeBSD__)
	sprintf (buf, "%s.%d", dev->prefix, epnum & 0x0f);
#else
	sprintf (buf, "%s.%02d", dev->prefix, epnum & 0x0f);
#endif
	fd = open (buf, flags, 0);
	if (fd < 0)
	{
		fprintf(stderr,	"pusb_endpoint_open: couldn't open endpoint %s: %s\n",
		  buf, strerror(errno));
		return (0);
	}
	
	return (make_endpoint(fd));
}

int pusb_endpoint_close (pusb_endpoint_t ep)
{
	int ret;
	
	ret = close (ep->fd);
	free (ep);
	
	return (ret);
}

int pusb_endpoint_get_descriptor (pusb_endpoint_t ep)
{
  if (!ep)
    return (-1);

  return (ep->fd);
}

int pusb_endpoint_read_bulk (pusb_endpoint_t ep, unsigned char *buf, int size,
		int timeout)
{
	int one = 1;
	int readed = 0;
	int ret;
	
	ret = ioctl(ep->fd, USB_SET_TIMEOUT, &timeout);
	if (ret < 0)
		return (-1);
	
	ret = ioctl (ep->fd, USB_SET_SHORT_XFER, &one);
    if (ret < 0)
		return (-1);
	
	do
	{
		ret = read (ep->fd, buf + readed, size - readed);
		if (ret < 0)
			return (-1);
		
		readed += ret;
	}
	while ((ret > 0) && (readed < size));

	return (readed);
}

int pusb_endpoint_write_bulk (pusb_endpoint_t ep, const unsigned char *buf,
		int size, int timeout)
{
	int sent = 0;
	int ret;
	
//	ret = ioctl(ep->fd, USB_SET_TIMEOUT, &timeout);
//	if (ret < 0)
//		return (-1);

	do
	{
	 	ret = write (ep->fd, buf + sent, size - sent);
		if (ret < 0)
			return (-1);
		
		sent += ret;
	}
	while ((ret > 0) && (sent < size));
			
	return (sent);
}

int pusb_endpoint_read_int (pusb_endpoint_t ep, unsigned char *buf, int size)
{
	int ret;

	ret = read (ep->fd, buf, size);
	if (ret < 0)
		return (-1);
		
	return (ret);
}

int pusb_endpoint_read_iso (pusb_endpoint_t ep, unsigned char *buf, int size)
{
	int ret;
	
	ret = read (ep->fd, buf, size);
	if (ret < 0)
		return (-1);

	return (ret);
}
